/*
  BeNoPPP
  Memory Block
  (c) 2003 - Sylvain Tertois
*/

#include "MemoryBlock.h"
#include <stdio.h>
#include <errno.h>
#include <ctype.h>

// constructor
MemoryBlock::MemoryBlock()
{
	m_CurrentBank=INSTRUCTION;
}

// destructor
MemoryBlock::~MemoryBlock()
{
}

// choose what memory bank to use
void MemoryBlock::SetMemoryBank(MemoryBank select)
{
	m_CurrentBank=select;
}

// returns the bank size, in words
unsigned int MemoryBlock::GetBankSize()
{
	return GetBank()->BufferLength()/2;
}

// go to beginning of bank
void MemoryBlock::ResetPosition()
{
	GetBank()->Seek(0,SEEK_SET);
}

// read a word
unsigned int MemoryBlock::ReadWord()
{
	short l_Word;
	GetBank()->Read(&l_Word,2);
	return l_Word;
}

// write a word
void MemoryBlock::WriteWord( unsigned int word)
{
	short l_Word(word);
	GetBank()->Write(&l_Word,2);
}

// go to a certain address
void MemoryBlock::GoToAddress( unsigned int address)
{
	if (address>=0x2100)
	{
		// we are in the data bank
		SetMemoryBank(DATA);
		address-=0x2100;
	}
	else if (address>=0x2000)
	{
		// we are in the configuration bank
		SetMemoryBank(CONFIGURATION);
		address-=0x2000;
	}
	else
		// we are in the instruction bank
		SetMemoryBank(INSTRUCTION);
	
	// seek to position
	GetBank()->Seek(address*2,SEEK_SET);
}

// give the name as a string of the given format
const char *MemoryBlock::GetFormatName(HexFormat format)
{
	switch(format)
	{
		case INHX8M:
			return "INHX8M";
		
		case INHX16:
			return "INHX16";
		
		case INHX32:
			return "INHX32";
		
		default:
			return "unknown";
	}
}

// start a new line
void MemoryBlock::FW_StartLine(FILE *file, int &checksum)
{
	// start with a ':'
	if (fprintf(file,":")<0)
		throw(errno);
		
	// reset the checksum
	checksum=0;
}

// write a byte
void MemoryBlock::FW_WriteInt8(FILE *file, char byte, int &checksum)
{
	if (fprintf(file,"%02X",((int)byte)&0xff)<0)
		throw(errno);
	checksum += byte;			// update checksum
}

// write a 16 bits word, low endian
void MemoryBlock::FW_WriteInt16L(FILE *file, short word, int &checksum)
{
	FW_WriteInt8(file,word & 0xff,checksum);
	FW_WriteInt8(file,word >> 8,checksum);
}

// write a 16 bits word, big endian
void MemoryBlock::FW_WriteInt16B(FILE *file, short word, int &checksum)
{
	FW_WriteInt8(file,word >> 8,checksum);
	FW_WriteInt8(file,word & 0xff,checksum);
}

// end of line
void MemoryBlock::FW_EndLine(FILE *file, int &checksum)
{
	// write a bute so that the checksum is 0
	FW_WriteInt8(file, (-checksum)&0xff,checksum);
	
	// new line
	if (fprintf(file,"\n")<0)
		throw(errno);
}

// write a full line
void MemoryBlock::FW_WriteLine( FILE *file, unsigned int address, unsigned int length, HexFormat format)
{
	int l_Checksum;
	FW_StartLine(file, l_Checksum);									// start the line
	switch(format)
	{
		case INHX8M:
		case INHX32:														// INHX8 and INHX32 formats
			FW_WriteInt8(file, (char)(2*length), l_Checksum);			// write line length
			FW_WriteInt16B(file, (short)(2*address), l_Checksum);	// write address
			FW_WriteInt8(file, 0, l_Checksum);							// write a 0
			while((length--)>0)
				FW_WriteInt16L(file, ReadWord(), l_Checksum);		// write the words, little endian mode
			break;

		case INHX16:														// INHX16 formats
			FW_WriteInt8(file, length, l_Checksum);					// write line length
			FW_WriteInt16B(file, address, l_Checksum);				// write address
			FW_WriteInt8(file, 0, l_Checksum);							// write a 0
			while((length--)>0)
				FW_WriteInt16B(file, ReadWord(), l_Checksum);		// write the words, big endian mode
			break;

		default:
			throw(-1);														// unknown format
	}
	FW_EndLine(file,l_Checksum);										// end of line
}

// read 4 bits (one hex digit)
int MemoryBlock::FR_ReadInt4(const char *&ptr)
{
	// read a character
	char c=*ptr++;
	
	// is it a decimal digit?
	if ((c>='0') && (c<='9'))
		return c-'0';

	// is it an uppercase letter?
	if ((c>='A') && (c<='F'))
		return c-'A'+0xA;
		
	// is it a lowercase letter?
	if ((c>='a') && (c<='f'))
		return c-'a'+0xA;

	// unkown character
	throw(-1);
}

// read a byte
int MemoryBlock::FR_ReadInt8(const char *&ptr, int &checksum)
{
	int l_Byte=(FR_ReadInt4(ptr)<<4)+FR_ReadInt4(ptr);

	// update the checksum
	checksum+=l_Byte;
	
	return l_Byte;
}

// read a 16 bits word, little endian mode
int MemoryBlock::FR_ReadInt16L(const char *&ptr, int &checksum)
{
	return FR_ReadInt8(ptr,checksum)+(FR_ReadInt8(ptr,checksum)<<8);
}

// read a 16 bits word, big endian mode
int MemoryBlock::FR_ReadInt16B(const char *&ptr, int &checksum)
{
	return (FR_ReadInt8(ptr,checksum)<<8)+FR_ReadInt8(ptr,checksum);
}

// write the memoryblock to an hex file
int MemoryBlock::WriteToFile(const char *filename, HexFormat format)
{
	FILE *l_File = NULL;
	try		// an exception is raised if there is an I/O error
	{
		int l_Checksum;
		
		// create or open the file for writing
		l_File = fopen(filename,"wt");
		if (l_File==NULL)
			throw(errno);
		
		// if the format is INHX32, we must first say that we are at address 0
		if (format == INHX32)
		{
			FW_StartLine(l_File,l_Checksum);
			FW_WriteInt8(l_File,2,l_Checksum);
			FW_WriteInt16B(l_File,0,l_Checksum);
			FW_WriteInt8(l_File,4,l_Checksum);
			FW_WriteInt16B(l_File,0,l_Checksum);
			FW_EndLine(l_File,l_Checksum);
		}
		
		// loop through the banks
		MemoryBank l_MemoryBank=INSTRUCTION;
		bool l_Loop=true;
		while (l_Loop)
		{
			// bank start address
			unsigned int l_Address;
			switch(l_MemoryBank)
			{
				case INSTRUCTION:
					l_Address=0;
					break;
				
				case CONFIGURATION:
					l_Address=0x2000;
					break;
				
				case DATA:
					l_Address=0x2100;
					break;
					
				default:
					throw(-1);
			}
			
			// go to beginning of bank
			SetMemoryBank(l_MemoryBank);
			ResetPosition();
			
			// number of words to write
			unsigned int l_LeftToWrite=GetBankSize();
			
			while (l_LeftToWrite>0)
			{
				// line length is 8 words, less if there are less words to write
				unsigned int l_LineLength=8;
				if (l_LeftToWrite<l_LineLength)
					l_LineLength=l_LeftToWrite;
				
				// write the line
				FW_WriteLine(l_File,l_Address,l_LineLength, format);
				
				// decrease the number of words left to write, and increase the address
				l_LeftToWrite-=l_LineLength;
				l_Address+=l_LineLength;
			}
			
			// go to next memory bank	
			switch(l_MemoryBank)
			{
				case INSTRUCTION:
					l_MemoryBank=CONFIGURATION;
					break;
					
				case CONFIGURATION:
					l_MemoryBank=DATA;
					break;
					
				case DATA:
				default:
					l_Loop=false;
			}
		}
		
		// write the last line
		FW_StartLine(l_File,l_Checksum);
		FW_WriteInt8(l_File,0,l_Checksum);
		FW_WriteInt16B(l_File,0,l_Checksum);
		FW_WriteInt8(l_File,1,l_Checksum);
		FW_EndLine(l_File,l_Checksum);

		// close the file
		if (fclose(l_File)<0)
		{
			l_File=NULL;
			throw(errno);
		}
	}
	catch (int error)
	{
		// I/O error
		
		// close the file
		if (l_File != NULL)
			fclose(l_File);
			
		// return error code
		return error;
	}
	return 0;
}

// read hex file and put the contents in the memoryblock
int MemoryBlock::ReadFromFile(const char *filename)
{
	FILE *l_File = NULL;
	try		// an exception is raised if there is an I/O error
	{
		// guess the file format. We'll first suppose that it's INHX8M
		HexFormat l_Format=INHX8M;
		const unsigned int c_MaxSize=200;		// max line size (a real line should be way shorter than that)
		char l_Line[c_MaxSize+1];					// allocate a buffer for the line
		
		// open the file
		l_File = fopen(filename,"rt");
		if (l_File==NULL)
			throw(errno);
			
		// read one line
		while(fgets(l_Line,c_MaxSize,l_File))
		{
			const char *l_Pointer=l_Line;			// first character should be a ':'
			if (*l_Pointer++ != ':')					// no... this must be a comment line. skip line
				continue;
			
			int l_Checksum=0;
			// line length
			unsigned int l_LineLength=FR_ReadInt8(l_Pointer,l_Checksum);
			if (l_LineLength==0)
				break;									// this is the end of the file
				
			// address
			unsigned int l_Address=FR_ReadInt16B(l_Pointer,l_Checksum);

			if (l_Format != INHX16)
			{
				l_LineLength/=2;
				l_Address/=2;
			}

			// line type
			int l_Type=FR_ReadInt8(l_Pointer,l_Checksum);
			if (l_Type==4)
				// this file is in intel hx32 format
				l_Format=INHX32;
			else
			{
				// read the line
				GoToAddress(l_Address);
				if (l_Format==INHX16)
					for (unsigned int i_Word=0; i_Word<l_LineLength;i_Word++)
						WriteWord(FR_ReadInt16B(l_Pointer,l_Checksum));
				else // INHX8M or INHX32
					for (unsigned int i_Word=0; i_Word<l_LineLength;i_Word++)
						WriteWord(FR_ReadInt16L(l_Pointer,l_Checksum));
					
				// read the checksum
				FR_ReadInt8(l_Pointer,l_Checksum);
				
				// we should now be at the end of the line
				if ((*l_Pointer!=0) && (!isspace(*l_Pointer)))
				{
					if (l_Format==INHX8M)
					{
						// this means that this file is in INHX16 format, not INHX8M
						l_Format=INHX16;
						// return to beginning of file
						fseek(l_File,0,SEEK_SET);
					}
					else
						// file format error
						throw(-1);
				}
				
				// check the checksum
				if ((l_Checksum&0xff)!=0)
					throw(-1);
			}
			
		}
		
		// close the file
		if (fclose(l_File)<0)
		{
			l_File=NULL;
			throw(errno);
		}
	}
	catch (int error)
	{
		// there has been an I/O error
		// close the file
		if (l_File != NULL)
			fclose(l_File);
		
		// return the error code
		return error;
	}
	
	// reset all banks positions
	SetMemoryBank(INSTRUCTION);
	ResetPosition();
	SetMemoryBank(CONFIGURATION);
	ResetPosition();
	SetMemoryBank(DATA);
	ResetPosition();
	
	return 0;
}
