/*
  BeNoPPP
  Memory Block
  (c) 2003 - Sylvain Tertois
*/

#ifndef MEMORYBLOCK_H
#define MEMORYBLOCK_H

#include <DataIO.h>

// a memory block represents all the memory in a PIC. It is divided in three banks:
// 1) instruction memory (14 bits words, addresses 0x0000-0x1FFF)
// 2) configuration memory, including the ID words and the configuration word (14 bit words addresses 0x2000-0x2007)
// 3) data memory (8 bits words, addresses 0x2100-0x3FFF)
// in this object, every stored word is 16 bits long, and only its 14 or 8 LSB are used.

// memory banks
typedef enum _memorybank {
	INSTRUCTION,
	CONFIGURATION,
	DATA
} MemoryBank;

// file formats
typedef enum _hexformat {
	INHX8M=0,
	INHX16,
	INHX32,
	NUMFORMATS
} HexFormat;

typedef struct _IO_FILE FILE;

// Memoryblock class
class MemoryBlock
{
public:
	MemoryBlock();
	~MemoryBlock();
	
	// read/write in memory functions
	void SetMemoryBank(MemoryBank select);		// choose what memory bank to use
	unsigned int GetBankSize();							// returns the bank size, in words
	void ResetPosition();									// go to beginning of bank
	unsigned int ReadWord();							// read a word
	void WriteWord(unsigned int word);				// write a word
	void GoToAddress(unsigned int address);			// go to a certain address
	
	// read/write to/from file
	int WriteToFile(const char *filename, HexFormat format);
		// return 0 if ok, negative value if error
	int ReadFromFile(const char *filename);
		// return 0 if ok, negative value if error
	
	// utilities
	// give the name as a string of the given format
	static const char *GetFormatName(HexFormat format);
	
private:
	BMallocIO m_IMemory;	// instructions memory
	BMallocIO m_CMemory;	// configuration memory
	BMallocIO m_DMemory;	// data memory
	
	MemoryBank m_CurrentBank;
	
	// utilities for file writing
	void FW_StartLine(FILE *file, int &checksum);	// start a new line and resets the checksum
	void FW_WriteInt8(FILE *file, char byte, int &checksum);	// write a byte
	void FW_WriteInt16L(FILE *file, short word, int &checksum);	// write a word, little endian
	void FW_WriteInt16B(FILE *file, short word, int &checksum);	// write a word, big endian
	void FW_EndLine(FILE *file, int &checksum);	// end the line and write the checksum
	void FW_WriteLine(FILE *file, unsigned int address, unsigned int length, HexFormat format);

	// utilities for file reading
	int FR_ReadInt4(const char *&ptr);						// read a digit
	int FR_ReadInt8(const char *&ptr, int &checksum);	// read a byte
	int FR_ReadInt16L(const char *&ptr, int &checksum);	// read a word, little endian
	int FR_ReadInt16B(const char *&ptr, int &checksum);	// read a word, big endian
			
	inline BMallocIO *GetBank() 
	{
		switch(m_CurrentBank)
		{
			case INSTRUCTION:
				return &m_IMemory;
			
			case CONFIGURATION:
				return &m_CMemory;

			case DATA:
			default:
				return &m_DMemory;
		}
	}
};

#endif //MEMORYBLOCK_H
