/*
  BeNoPPP
  Driver interface
  (c) 2003 - Sylvain Tertois
  Based on NoPPP for Linux 0.3 by Michael A. Covington, Claus Fuetterer and Geoff McCaughan
*/

#include "NoPPPDriver.h"
#include "ControlCodes.h"
#include <fcntl.h>
#include <sys/ioctl.h>
#include <kernel/OS.h>
#include <errno.h>

// constructor
NoPPPDriver::NoPPPDriver(unsigned int ioBase)
{
	// open the driver
	m_FileDescriptor = open("/dev/misc/noppp",O_RDWR);
	
	if (m_FileDescriptor>0)
	{
		// no error while opening it. Trying to send a command (set the IO base)
		status_t l_Status = ioctl(m_FileDescriptor,BNPPP_SETBASEPORT,&ioBase);
		if (l_Status != B_OK)
		{
			// error. Is status is -1, the error number is in errno
			if (l_Status==-1)
				l_Status=errno;
				
			// throw an exception
			throw l_Status;
		}
	}
	else
		// error while opening the driver
		throw ((status_t)m_FileDescriptor);
}

NoPPPDriver::~NoPPPDriver()
{
	// close the driver if it was opened
	if (m_FileDescriptor>0)
		close(m_FileDescriptor);
}

// for the next functions: first send the command to the driver, and throw an exception if an error occurs
// if ioctl returns -1, it means that the real error code is in errno

// call this before sending a data bit to the PIC
void NoPPPDriver::DataWritable()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_DATAWRITABLE);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// call this before reading a data bit from the PIC
void NoPPPDriver::DataReadable()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_DATAREADABLE);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// changes the bit to write to 0
void NoPPPDriver::DataDown()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_DATADOWN);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// changes the bit to write to 1
void NoPPPDriver::DataUp()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_DATAUP);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// changes the clock to 0
void NoPPPDriver::ClockDown()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_CLOCKDOWN);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// changes the clock to 1
void NoPPPDriver::ClockUp()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_CLOCKUP);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// enables vpp (-> pic enters programing mode)
void NoPPPDriver::VppOn()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_VPPON);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// disables vpp (-> pic returns to normal mode)
void NoPPPDriver::VppOff()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_VPPOFF);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// enables vpp (-> pic power on)
void NoPPPDriver::VddOn()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_VDDON);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// disables vpp (-> pic power off)
void NoPPPDriver::VddOff()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_VDDOFF);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// read 1 bit
bool NoPPPDriver::ReadData()
{
	char l_DataByte;
	status_t l_Status;
	l_Status=ioctl(m_FileDescriptor,BNPPP_DATAIN,&l_DataByte);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}

	return (bool)l_DataByte;
}

// vpp off, data down, clock down, data readable
void NoPPPDriver::AllPinsLow()
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_ALLPINSLOW);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// set the pins to use
void NoPPPDriver::SetPins(UsePin *pins)
{
	status_t l_Status = ioctl(m_FileDescriptor,BNPPP_SETPINS,pins);
	if (l_Status != B_OK)
	{
		if (l_Status==-1)
			l_Status=errno;
		throw l_Status;
	}
}

// returns true if the programer is detected
bool NoPPPDriver::DetectHardware()
{
	DataWritable();	// SLCTIN down
	DataUp();			// AUTOFD up	
	snooze(10);		// wait 10µs
	
	if (ReadData())	// should read 0
		return false;
		
	DataReadable();	// SLCTIN up
	DataUp();			// AUTOFD up
	snooze(10);		// wait 10µs
	
	if (!ReadData())	// should read 1
		return false;
	
	return true;
}

// send one bit to the PIC (handles clock and data, you must call DataWritable() before)
void NoPPPDriver::SendBit(bool data)
{
	// but the data line to the correct level
	if (data)
		DataUp();
	else
		DataDown();
		
	ClockUp();		// rising edge	
	snooze(1);		// wait tset1
	
	ClockDown();	// data is clocked into PIC on this falling edge

	snooze(1);		// wait thld1
	
	DataDown();	// idle with data line low
}

// receive one bit from the PIC (handles clock and data, you must call DataReadable() before)
bool NoPPPDriver::ReceiveBit()
{
	bool l_Data;
	
	ClockUp();		// rising edge
	snooze(1);		// wait tdly3
	
	ClockDown();	// data is ready just before this
	l_Data = ReadData();
	
	snooze(1);
	
	return l_Data;
}

// send 6 bits, lsb first (you don't need to call DataWritable() before)
void NoPPPDriver::SendCommand( PicCommand command)
{
	int l_Command(command);
	
	DataWritable();							// prepare to write
	snooze(2);									// wait thld0
	for (unsigned int i=0; i<6; i++)			// loop for the 6 bits
	{
		SendBit((bool)(l_Command & 1));	// send the lsb bit
		l_Command >>= 1;					// go to following bit
	}
	snooze(2);									// wait tdy2
}

// send 14 bits, lsb first (you don't need to call DataWritable() before)
void NoPPPDriver::SendData(unsigned int data)
{
	DataWritable();							// prepare to write
	snooze(2);									// wait thld0
	SendBit(0);								// one garbage bit
	for (unsigned int i=0; i<14; i++)		// loop for the 14 bits
	{
		SendBit((bool)(data & 1));			// send the lsb bit
		data >>= 1;							// go to following bit
	}
	SendBit(0);								// one garbage bit
	snooze(2);									// wait tdy2	
}

// send 14 bits, lsb first (you don't need to call DataReadable() before)
unsigned int NoPPPDriver::ReceiveData()
{
	unsigned int l_Data=0;					// this will hold the read data
	unsigned int l_Mask=1;					// this mask has one bit at 1, the one we are currently reading
	
	DataReadable();							// prepare to read
	snooze(2);									// wait thld0
	ReceiveBit();								// one garbage bit
	for (unsigned int i=0; i<14; i++)		// loop for the 14 bits
	{
		if (ReceiveBit())						// if the read bit is 1
			l_Data |= l_Mask;				// put the bit at 1 in l_Data
		l_Mask <<=1;						// and move the mask to the next bit
	}
	ReceiveBit();								// another garbage bit
	snooze(2);									//wait tdly2
	
	return l_Data;								// and return the data
}
