/*
  BeNoPPP
  Read Thread
  (c) 2003 - Sylvain Tertois
*/

#include "ReadThread.h"
#include "ReadView.h"
#include "PicList.h"
#include "NoPPPDriver.h"
#include "BeNoPPPApp.h"
#include "MemoryBlock.h"

#include <Messenger.h>
#include <Alert.h>

// constructor
ReadThread::ReadThread( BMessenger *readview, unsigned int picType, HexFormat fileFormat, const char *fileName, unsigned int iobase, UsePin *pins, bool detect)
	: BLooper("Read Thread")
{
	m_ReadView=readview;
	m_PicType=picType;
	m_FileFormat=fileFormat;
	m_FileName=fileName;
	m_IOBase=iobase;
	m_Pins=pins;
	m_DetectHardware=detect;
	
	Run();
	PostMessage(RT_GO);
}

// destructor
ReadThread::~ReadThread()
{
	delete m_ReadView;
}

// hook function called when a message arrives
void ReadThread::MessageReceived( BMessage *message)
{
	switch (message->what)
	{
		case RT_GO:
			// start the work
			Go();
			break;
		
		default:
			BLooper::MessageReceived(message);
	}
}

// called by Go() with the status bar needs an update
void ReadThread::Update(const char *text)
{
	// construct the message
	BMessage l_Message(ReadView::RV_STATUSUPDATE);
	
	// add the text, if present
	if (text != NULL)
		l_Message.AddString("text",text);
	
	// send the message to the view
	m_ReadView->SendMessage(&l_Message);
}

// actual work
void ReadThread::Go()
{
	try		// an exception will be raised if an arror occurs
	{
		// open the driver and allocate a memory block to put the contents
		NoPPPDriver l_Driver(m_IOBase);
		l_Driver.SetPins(m_Pins);
		MemoryBlock l_Memory;
	
		// count how many update messages will be sent
		BMessage l_UpdateSizeMessage(ReadView::RV_STATUSMAX);
		l_UpdateSizeMessage.AddInt32("max",
			c_PicList[m_PicType].psize+4+1+c_PicList[m_PicType].dsize+4);
			
		// send the count to the view with the status bar
		m_ReadView->SendMessage(&l_UpdateSizeMessage);
		
		// initialisation
		l_Driver.AllPinsLow();
		m_ReadView->SendMessage(ReadView::RV_READSTARTED);
		
		// is the hardware present?
		if (m_DetectHardware && !l_Driver.DetectHardware())
		{
			// no
			unsigned int button = (new BAlert("No hardware","The programer hasn't been detected",
				"Proceed Anyway","Cancel",NULL,B_WIDTH_AS_USUAL,
				B_WARNING_ALERT))->Go();
			if (button==1)
			{
				// cancel
				l_Driver.AllPinsLow();
				throw int(0);
			}
		}
		l_Driver.AllPinsLow();
		
		// wait that the user inserts the PIC
		if ((new BAlert("Insert PIC","Insert the PIC in the programer","OK","Cancel",NULL))->Go() == 1)
			throw int(0);
		
		// prepare before turning on VPP (25ms wait before and after, to have clean signals while the PIC enters programing mode)
		l_Driver.VddOn();
		l_Driver.DataWritable();
		l_Driver.DataDown();
		l_Driver.ClockDown();
		snooze(25000);
		m_ReadView->SendMessage(ReadView::RV_VPPON);
		l_Driver.VppOn();
		snooze(25000);
		
		// READ
		// program memory
		Update("reading program memory");
		l_Memory.SetMemoryBank(INSTRUCTION);			// switch to the correct bank
		for (unsigned int i_Add=0; i_Add<c_PicList[m_PicType].psize; i_Add++)
		{
			l_Driver.SendCommand(READPROGRAM);		// send a read command
			l_Memory.WriteWord(l_Driver.ReceiveData());	// receive the data and put it in memory
			l_Driver.SendCommand(INCREMENTADD);		// go to next address
			Update();												// update the status bar
		}
		
		// configuration memory (ID)
		Update("reading id memory");
		l_Memory.SetMemoryBank(CONFIGURATION);		// switch to the correct bank
		l_Driver.SendCommand(LOADCONFIG);				// turn the PIC into config mode
		l_Driver.SendData(c_PicList[m_PicType].defaultConfig);
				// load default config (is this really good?)
		for (unsigned int i_Add=0; i_Add<4; i_Add++)
		{
			l_Driver.SendCommand(READPROGRAM);		// send a read command
			l_Memory.WriteWord(l_Driver.ReceiveData() & c_PicList[m_PicType].imask);
					// receive the data and put it in memory
			l_Driver.SendCommand(INCREMENTADD);		// go to next address
			Update();												// update the status bar
		}
		
		// goto config word (0x2007)
		// 3 jumps
		for (unsigned int i_Jump=0; i_Jump<3; i_Jump++)
		{
			l_Memory.WriteWord(0);
			l_Driver.SendCommand(INCREMENTADD);
		}
		
		// read config word
		Update("reading configuration word");
		l_Driver.SendCommand(READPROGRAM);		// send a read command
		l_Memory.WriteWord(l_Driver.ReceiveData() & c_PicList[m_PicType].cmask);
					// receive the data and put it in memory
		l_Driver.SendCommand(INCREMENTADD);		// go to next address
		Update();												// update the status bar

		// reset the PIC to read data memory
		l_Driver.VppOff();										// turn VPP off
		l_Driver.DataWritable();								// put all
		l_Driver.DataDown();									//  pins
		l_Driver.ClockDown();								//  down
		snooze(25000);										// wait 25ms
		l_Driver.VppOn();										// turn VPP on again
		snooze(25000);										// wait 25ms
		
		// read data memory
		l_Memory.SetMemoryBank(DATA);					// switch to the correct bank
		Update("reading data memory");
		for (unsigned int i_Add=0; i_Add<c_PicList[m_PicType].dsize; i_Add++)
		{
			l_Driver.SendCommand(READDATA);			// send a read command
			l_Memory.WriteWord(l_Driver.ReceiveData() & 0xFF);
					// receive the data and put it in memory
			l_Driver.SendCommand(INCREMENTADD);	// go to next address
			Update();
		}
				
		// finished
		l_Driver.VppOff();									// turn VPP off
		l_Driver.DataWritable();							// put
		l_Driver.DataDown();								// all pins
		l_Driver.ClockDown();							// down
		l_Driver.VddOff();
		
		// send a message to the view saying that it is finished
		m_ReadView->SendMessage(ReadView::RV_READSTOPPED);
		
		// write the memory contents to a hex file
		if (l_Memory.WriteToFile(m_FileName, m_FileFormat) < 0)
			(new BAlert("Write file error","Error while writing hex file","Cancel",NULL,NULL,
				B_WIDTH_AS_USUAL,B_STOP_ALERT))->Go();
		else
			(new BAlert("Reading finished","Reading finished","OK"))->Go();
		
		// this is the end of this thread
		PostMessage(B_QUIT_REQUESTED);
	}
	catch (status_t error)
	{
		// driver error
		ErrorAlert(error);									// display an error message
		m_ReadView->SendMessage(ReadView::RV_READSTOPPED);
				// say to the view that the work is finished
		PostMessage(B_QUIT_REQUESTED);			// stop the thread
	}
	catch (int i)
	{
		// user cancel
		m_ReadView->SendMessage(ReadView::RV_READSTOPPED);		
				// say to the view that the work is finished
		PostMessage(B_QUIT_REQUESTED);			// stop the thread
	}	
}
