/* ++++++++++
	driver.c
	Driver for the No-Parts Pic Programmer
	(c) 2003 - Sylvain Tertois
	Based on NoPPP for Linux 0.3 by Michael A. Covington, Claus Fuetterer and Geoff McCaughan
+++++ */

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ISA.h>

#include "ControlCodes.h"
#include "driver.h"

// Pin information
PinInfo g_PinInfo[NUMPINS]={
//		ad		bitnum	inv		// name
	{	0,		0,		0 },	// P_NONE*
	// parallel port - outputs
	{	2,		0,		1 },		// P_STROBE
	{	0,		0,		0 },	// P_D0
	{	0,		1,		0 },	// P_D1
	{	0,		2,		0 },	// P_D2
	{	0,		3,		0 },	// P_D3
	{	0,		4,		0 },	// P_D4
	{	0,		5,		0 },	// P_D5
	{	0,		6,		0 },	// P_D6
	{	0,		7,		0 },	// P_D7
	{	2,		1,		1 },		// P_AUTOFD
	{	2,		2,		0 },	// P_INIT
	{	2,		3,		1 },		// P_SELIN
	// - inputs
	{	1,		6,		0 },	// P_ACK
	{	1,		7,		1 },		// P_BUSY
	{	1,		5,		0 },	// P_PERROR
	{	1,		4,		0 },	// P_SEL
	{	1,		3,		0 },	// P_ERROR
	// serial ports - outputs
	{	3,		6,		0 },	// P_TXD
	{	4,		0,		0 },	// P_DTR
	{	4,		1,		0 },	// P_RTS
	// - inputs
	{	6,		4,		0 },	// P_CTS
	{	6,		7,		0 },	// P_CD
	{	6,		5,		0 },	// P_DSR
	{	6,		6,		0 }		// P_RI
};
									// names ending by '*' are handled specially
									
// default configuration (noppp)
UsePin		g_DefaultConfig[NUMFUNCTIONS] = {
//		pin				inv			// name
	{	P_AUTOFD,	0		},	// F_WRITE
	{	P_STROBE,	0		},	// F_CLK
	{	P_D0,			0		},	// F_MCLR
	{	P_NONE,		0		},	// F_PGM
	{	P_BUSY,		0		},	// F_READ
	{	P_NONE,		0		},	// F_VDD
	{	P_SELIN,		0		},	// F_IN1
	{	P_AUTOFD,	0		}	// F_IN2
};

/* ----------
	init_hardware - called once the first time the driver is loaded
----- */

isa_module_info *isa;

status_t
init_hardware (void)
{
	return B_OK;
}


/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
status_t
init_driver (void)
{
	// get the ISA module
	if (get_module(B_ISA_MODULE_NAME, (module_info**)&isa) < 0)
		return ENOSYS;
	return B_OK;
}


/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
void
uninit_driver (void)
{
	// free the ISA module
	put_module(B_ISA_MODULE_NAME);
}

	
/* ----------
	my_device_open - handle open() calls
----- */

static status_t
my_device_open (const char *name, uint32 flags, void** cookie)
{
	unsigned int i_ByteNum;
	
	DriverInfo *myInfo = (DriverInfo*)malloc(sizeof(DriverInfo));
	if (myInfo == NULL)
		return B_NO_MEMORY;
	else
	{
		myInfo->portNumber = 0x378;
		for (i_ByteNum=0; i_ByteNum<8; i_ByteNum++)
			myInfo->IOBytes[i_ByteNum]=0;
		CopyPinsTable(myInfo,g_DefaultConfig);
					
		*((DriverInfo**)cookie)=myInfo;
		return B_OK;
	}
}


/* ----------
	my_device_read - handle read() calls
----- */

static status_t
my_device_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	*num_bytes = 0;				/* tell caller nothing was read */
	return B_IO_ERROR;
}


/* ----------
	my_device_write - handle write() calls
----- */

static status_t
my_device_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	*num_bytes = 0;				/* tell caller nothing was written */
	return B_IO_ERROR;
}


/* 
/* ----------
	my_device_control - handle ioctl calls
----- */

#define COOKIE ((DriverInfo*)cookie)

static status_t
my_device_control (void* cookie, uint32 op, void* arg, size_t len)
{
	switch (op)
	{
		case BNPPP_SETBASEPORT:
		{
			int l_PortNum = *((int*)arg);
			
			// check that the requested port number is a LPT/COM port
			switch(l_PortNum)
			{
				case 0x378:
				case 0x278:
				case 0x288:
				case 0x388:
				case 0x3F8:
				case 0x2F8:
				case 0x3E8:
				case 0x2E8:
				
					COOKIE->portNumber=l_PortNum;
					break;
				
				default:
					return B_BAD_VALUE;
			}
					
			break;
		}

		case BNPPP_SETPINS:
			CopyPinsTable(COOKIE,(UsePin*)arg);
			break;
			
		case BNPPP_DATAWRITABLE:
			FunctionWrite(COOKIE,F_IN1,0);
			FunctionWrite(COOKIE,F_IN2,0);
			break;
		
		case BNPPP_DATAREADABLE:
			FunctionWrite(COOKIE,F_IN1,1);
			FunctionWrite(COOKIE,F_IN2,1);
			break;
		
		case BNPPP_DATADOWN:
			FunctionWrite(COOKIE,F_WRITE,0);
			break;

		case BNPPP_DATAUP:
			FunctionWrite(COOKIE,F_WRITE,1);
			break;

		case BNPPP_CLOCKDOWN:
			FunctionWrite(COOKIE,F_CLK,0);
			break;

		case BNPPP_CLOCKUP:
			FunctionWrite(COOKIE,F_CLK,1);
			break;
			
		case BNPPP_VPPON:
			FunctionWrite(COOKIE,F_PGM,1);
			FunctionWrite(COOKIE,F_MCLR,0);
			break;
			
		case BNPPP_VPPOFF:
			FunctionWrite(COOKIE,F_MCLR,1);
			FunctionWrite(COOKIE,F_PGM,0);
			break;
			
		case BNPPP_DATAIN:
		{
			*((uint8*)arg) = FunctionRead(COOKIE,F_READ);
			break;
		}
		
		case BNPPP_VDDON:
			FunctionWrite(COOKIE,F_VDD,1);
			break;
			
		case BNPPP_VDDOFF:
			FunctionWrite(COOKIE,F_VDD,0);
			break;

		case BNPPP_ALLPINSLOW:
			FunctionWrite(COOKIE,F_WRITE,0);
			FunctionWrite(COOKIE,F_CLK,0);
			FunctionWrite(COOKIE,F_MCLR,1);
			FunctionWrite(COOKIE,F_PGM,0);
			FunctionWrite(COOKIE,F_IN1,0);
			FunctionWrite(COOKIE,F_IN2,0);
			FunctionWrite(COOKIE,F_VDD,0);
			break;
		
		default:
			return B_BAD_VALUE;
	}
	
	return B_OK;
}


/* ----------
	my_device_close - handle close() calls
----- */

static status_t
my_device_close (void* cookie)
{
	free(cookie);
	return B_OK;
}


/* -----
	my_device_free - called after the last device is closed, and after
	all i/o is complete.
----- */
static status_t
my_device_free (void* cookie)
{
	return B_OK;
}


/* -----
	null-terminated array of device names supported by this driver
----- */

static const char *my_device_name[] = {
	"misc/noppp",
	NULL
};

/* -----
	function pointers for the device hooks entry points
----- */

device_hooks my_device_hooks = {
	my_device_open, 			/* -> open entry point */
	my_device_close, 			/* -> close entry point */
	my_device_free,			/* -> free cookie */
	my_device_control, 		/* -> control entry point */
	my_device_read,			/* -> read entry point */
	my_device_write			/* -> write entry point */
};

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

const char**
publish_devices()
{
	return my_device_name;
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

device_hooks*
find_device(const char* name)
{
	return &my_device_hooks;
}

// utility functions
// write
void FunctionWrite(DriverInfo *cookie, Function function, int bit)
{
	Pin l_PinNumber;
	unsigned int l_RegNum, l_Mask;
	
	// get the pin number
	l_PinNumber = cookie->usepins[function].pin;
	if (l_PinNumber==P_NONE)
		return;
	
	// get the register num
	l_RegNum = g_PinInfo[l_PinNumber].address;
	
	// create the mask
	l_Mask = 1 << g_PinInfo[l_PinNumber].bitnum;
	
	// invert the information if needed
	if (g_PinInfo[l_PinNumber].inverted)
		bit=(bit?0:1);
	if (cookie->usepins[function].inverted)
		bit=(bit?0:1);
	
	// change the IO byte
	if (bit)
		// put to 1
		cookie->IOBytes[l_RegNum] |= l_Mask;
	else
		// put to 0
		cookie->IOBytes[l_RegNum] &= ~l_Mask;
	
	// write the new IO byte
	isa->write_io_8(cookie->portNumber+l_RegNum,cookie->IOBytes[l_RegNum]);
}

// read
int FunctionRead(DriverInfo *cookie, Function function)
{
	Pin l_PinNumber;
	unsigned int l_RegNum, l_Mask,l_Bit;
	
	// get the pin number
	l_PinNumber = cookie->usepins[function].pin;
	if (l_PinNumber==P_NONE)
		return 0;
	
	// get the register num
	l_RegNum = g_PinInfo[l_PinNumber].address;
	
	// create the mask
	l_Mask = 1 << g_PinInfo[l_PinNumber].bitnum;

	// read the bit
	l_Bit = (isa->read_io_8(cookie->portNumber+l_RegNum) & l_Mask)?1:0;
	
	// invert the information if needed
	if (g_PinInfo[l_PinNumber].inverted)
		l_Bit=(l_Bit?0:1);
	if (cookie->usepins[function].inverted)
		l_Bit=(l_Bit?0:1);
	
	// return the read bit
	return l_Bit;
}

// copy a pins table into the driver cookie
void CopyPinsTable(DriverInfo *cookie, const UsePin *source)
{
	memcpy(cookie->usepins,source,NUMFUNCTIONS*sizeof(UsePin));
}

